#include <iostream>
#include <cassert>
#include "z3-api/c++/z3++.h"

class ssa_info
{
private:
    std::string base_name;
    unsigned num_copies;
public:
    explicit ssa_info(std::string name) : base_name(name), num_copies(0) {};
    z3::expr get(z3::context& ctx)
    {
        std::string name = base_name + "_" + std::to_string(num_copies);
        num_copies += 1;
        return ctx.bv_const(name.c_str(), 32);
    }
};

int check_sat(z3::solver& solver)
{
    auto res = solver.check();
    if(res == z3::unsat)
    {
        std::cout << "unsat" << std::endl;
        return 20;
    }
    else if(res == z3::sat)
    {
        auto model = solver.get_model();
        unsigned int num_consts = model.num_consts();
        for (unsigned int i = 0; i < num_consts; i++)
        {
            auto decl = model.get_const_decl(i);
            auto val = model.get_const_interp(decl).get_numeral_int();
            std::cout << decl.name() << " " << val << std::endl;
        }
        return 10;
    }
    return 30;
}


void my_sqrt()
{
    z3::context ctx;
    z3::solver solver(ctx);

    ssa_info input_info("input");
    z3::expr input = input_info.get(ctx);

    solver.add(input.extract(input.get_sort().bv_size() - 1, input.get_sort().bv_size() - 17) == 0);
    // solver.add(input == 100);

    // unsigned top = input;

    ssa_info top_info("top");
    z3::expr top = top_info.get(ctx);
    solver.add(top == input);

    // unsigned bot = 0;

    ssa_info bot_info("bot");
    z3::expr bot = bot_info.get(ctx);
    solver.add(bot == 0);

    // unsigned mid;

    ssa_info mid_info("mid");
    z3::expr mid(ctx);

    ssa_info mid2_info("mid2");

    for (int i = 0; i < 10; i++)
    {
        // mid = (top + bot) / 2;
        mid = mid_info.get(ctx);
        solver.add(mid == (top + bot) / 2);
        // unsigned mid2 = mid * mid;
        z3::expr mid2 = mid2_info.get(ctx);
        solver.add(mid2 == mid * mid);

        // if (mid2 < input)
        // { bot = mid; }
        // else
        // { bot = bot; }
        z3::expr cond = z3::ult(mid2, input);
        z3::expr bot_then = bot_info.get(ctx);
        solver.add(bot_then == mid);
        z3::expr bot_else = bot_info.get(ctx);
        solver.add(bot_else == bot);
        bot = bot_info.get(ctx);
        solver.add(bot == z3::ite(cond, bot_then, bot_else));
        // solver.add(z3::implies(cond, bot == bot_then));
        // solver.add(z3::implies(!cond, bot == bot_else));

        // if (mid2 > input)
        // { top = mid; }

        cond = z3::ugt(mid2, input);
        z3::expr top_then = top_info.get(ctx);
        solver.add(top_then == mid);
        z3::expr top_else = top_info.get(ctx);
        solver.add(top_else == top);
        top = top_info.get(ctx);
        solver.add(top == z3::ite(cond, top_then, top_else));
    }

    std::cout << mid << std::endl;

    // mid = (top + bot) / 2;
    mid = mid_info.get(ctx);
    solver.add(mid == (top + bot) / 2);

    // unsigned mip = mid + 1;
    ssa_info mip_info("mip");
    z3::expr mip = mip_info.get(ctx);
    solver.add(mip == mid + 1);

    // assert((mid * mid <= input) ||
    //           (mip * mip >= input));

    z3::expr assertion = (mid * mid <= input) && (mip * mip >= input);
    solver.add(!assertion);

    check_sat(solver);
}

int main(int argc, char* argv[])
{
    my_sqrt();

    return 0;
}


